import sys
sys.path.append('/home1/jrudoler/src/')
from retrieval_help import create_baseline_events, free_epochs
import cmlreaders as cml
import numpy as np
import pandas as pd
from ptsa.data.filters import ButterworthFilter, MorletWaveletFilter
from ptsa.data.concat import concat
from ptsa.data.timeseries import TimeSeries
from scipy.stats import zscore
from irasa.IRASA import IRASA
import CMLDask as da

def trim_axis(arr1, arr2, axis=-1):
    if (arr1.shape[axis] != arr2.shape[axis]):
        min_length = min([arr1.shape[axis], arr2.shape[axis]])
        arr1 = np.take(arr1, range(min_length), axis)
        arr2 = np.take(arr2, range(min_length), axis)
    return arr1, arr2

def trim_time_ts(arr1, arr2):
    if (len(arr1['time']) != len(arr2['time'])):
        min_length = min([len(arr1['time']), len(arr2['time'])])
        arr1 = np.take(arr1, range(min_length), -1)
        arr1 = arr1.assign_coords({'time':np.take(arr1['time'].values, range(min_length))})
        arr2 = np.take(arr2, range(min_length), -1)
        arr2 = arr2.assign_coords({'time':np.take(arr2['time'].values, range(min_length))})
    return arr1, arr2

def SSL_transform(x):
    # Shifted, Symmetric Log Transform to suppress extrema
    return np.sign(x) * np.log10(np.abs(x)+1)

def calc_bip_oscill(sub, settings_path='settings_IRASA_hmax_2.pkl'):   
    settings = da.Settings.Load(filename=settings_path)
    subjects = pd.read_pickle(settings.subj_pairs_path)
    hippo_pairs = subjects[subjects.Subject==sub].hippo_pairs.iloc[0]
    data = cml.get_data_index(kind = 'r1'); data = data[data['experiment'] =='FR1']; data = data[data['subject']==sub]
    recalls = []
    full_eeg = []
    # iterate then average over sessions
    stored_exceptions = []
    for i, row in data.iterrows():
        reader = cml.CMLReader(sub, 'FR1', session = row['session'], 
                               localization=row['localization'], montage=row['montage'])
        events = reader.load('task_events')
        word_events= events[(events.type == 'WORD')&(events.eegoffset != -1)]
        pairs = reader.load('pairs')
        if not np.isin(hippo_pairs, pairs.label).all():
            # localization does not match
            continue
        good_pairs = pairs[np.isin(pairs.label, hippo_pairs)]
        # resample so that number of frequencies is the same
        eeg = reader.load_eeg(word_events, scheme = good_pairs,
                              rel_start=settings.rel_start, rel_stop=settings.rel_stop)
        eeg = eeg.resample(500.)
        eeg = eeg.to_ptsa()
        if not len(full_eeg):
            full_eeg.append(eeg)
        else:
            for j in range(len(full_eeg)):
                eeg, full_eeg[j] = trim_time_ts(eeg, full_eeg[j])
            full_eeg.append(eeg)
        recalls.append(word_events)
    full_eeg = concat(full_eeg, 'event')
    events = pd.concat(recalls)
    bt_filt = ButterworthFilter(freq_range=[58., 62.])
    full_eeg = full_eeg.filter_with(bt_filt)
    irasa = IRASA(full_eeg, settings.freqs, samplerate = full_eeg.samplerate.values, hset=settings.hset)
    p, plaw = irasa.plaw_fit()
    osc = irasa.mixed - irasa.fractal
    # log transform
    osc = SSL_transform(osc)
    mixed = np.log10(irasa.mixed)
    fractal = np.log10(irasa.fractal)
    
    ts = TimeSeries(np.stack([osc, fractal, mixed]),
           dims=('type', 'event', 'channel', 'frequency'),
           coords = {'type':['oscillatory', 'fractal', 'mixed'],
                     'event':pd.MultiIndex.from_frame(events[["eegoffset","session", "item_name", "recalled"]]),
                     'frequency':settings.freqs,
                     'channel':full_eeg.channel.values, 'samplerate':full_eeg.samplerate.values}
          )
    ts = ts.groupby('session').reduce(zscore, dim='event', ddof=1)
    
    
    if np.mean(ts.recalled)<.08:
        # subject remembers less than one word per list
        raise Exception('Subject does not meet recall requirements')
    if settings.save:
        ts.to_hdf(settings.data_dir + f"{sub}_bip_irasa.h5")
    return ts

def calc_avg_oscill(sub, settings_path='settings_IRASA_hmax_2.pkl'):   
    settings = da.Settings.Load(filename=settings_path)
    subjects = pd.read_pickle(settings.subj_pairs_path)
    hippo_pairs = subjects[subjects.Subject==sub].hippo_pairs.iloc[0]
    hippo_pairs = np.ravel([i.split('-') for i in hippo_pairs])
    data = cml.get_data_index(kind = 'r1'); data = data[data['experiment'] =='FR1']; data = data[data['subject']==sub]
    recalls = []
    full_eeg = []
    # iterate then average over sessions
    stored_exceptions = []
    for i, row in data.iterrows():
        reader = cml.CMLReader(sub, 'FR1', session = row['session'], 
                               localization=row['localization'], montage=row['montage'])
        events = reader.load('task_events')
        word_events= events[(events.type == 'WORD')&(events.eegoffset != -1)]
        contacts = reader.load('contacts')
        if not np.isin(hippo_pairs, contacts.label).all():
            # localization does not match
            continue
        # resample so that number of frequencies is the same
        eeg = reader.load_eeg(word_events, scheme = contacts,
                              rel_start=settings.rel_start, rel_stop=settings.rel_stop)
        eeg = eeg.resample(500.)
        eeg = eeg.to_ptsa()
        # average reference
        good_contacts = np.isin(contacts.label, hippo_pairs)
        eeg = eeg[:, good_contacts, :] - eeg.mean('channel')
        if not len(full_eeg):
            full_eeg.append(eeg)
        else:
            for j in range(len(full_eeg)):
                eeg, full_eeg[j] = trim_time_ts(eeg, full_eeg[j])
            full_eeg.append(eeg)
        recalls.append(word_events)
    full_eeg = concat(full_eeg, 'event')
    events = pd.concat(recalls)
    bt_filt = ButterworthFilter(freq_range=[58., 62.])
    full_eeg = full_eeg.filter_with(bt_filt)
    irasa = IRASA(full_eeg, settings.freqs, samplerate = full_eeg.samplerate.values, hset=settings.hset)
    p, plaw = irasa.plaw_fit()
    osc = irasa.mixed - irasa.fractal
    # log transform
    osc = SSL_transform(osc)
    mixed = np.log10(irasa.mixed)
    fractal = np.log10(irasa.fractal)
    
    ts = TimeSeries(np.stack([osc, fractal, mixed]),
           dims=('type', 'event', 'channel', 'frequency'),
           coords = {'type':['oscillatory', 'fractal', 'mixed'],
                     'event':pd.MultiIndex.from_frame(events[["eegoffset","session", "item_name", "recalled"]]),
                     'frequency':settings.freqs,
                     'channel':full_eeg.channel.values, 'samplerate':full_eeg.samplerate.values}
          )
    ts = ts.groupby('session').reduce(zscore, dim='event', ddof=1)
    
    
    if np.mean(ts.recalled)<.08:
        # subject remembers less than one word per list
        raise Exception('Subject does not meet recall requirements')
    if settings.save:
        ts.to_hdf(settings.data_dir + f"{sub}_avg_irasa.h5")
    return ts

## retrieval
def calc_bip_oscill_rec(sub, settings_path='settings_IRASA_retrieval_hmax_2.pkl'):
    import sys
    sys.path.append('/home1/jrudoler/src/')
    from retrieval_help import create_baseline_events, free_epochs
    settings = da.Settings.Load(filename=settings_path)
    subjects = pd.read_pickle(settings.subj_pairs_path)
    hippo_pairs = subjects[subjects.Subject==sub].hippo_pairs.iloc[0]
    data = cml.get_data_index(kind = 'r1'); data = data[data['experiment'] =='FR1']; data = data[data['subject']==sub]
    full_eeg = []
    cnt = 0
    events = []
    for i, row in data.iterrows(): # iterate then average over sessions
        reader = cml.CMLReader(sub, 'FR1', session = row['session'], 
                               localization=row['localization'], montage=row['montage'])
        evs = reader.load('all_events')
        evs = evs[evs.eegoffset>0]
        # drop retrieval events where preceding word was < 1000 ms before
        rec_mask = (evs.type=='REC_WORD').values.astype(bool)
        evs = evs.drop(index=evs[rec_mask].index[(evs[rec_mask].mstime.diff()<1000)].values)
        # get matched baseline events to compare to recalls
        try:
            evs = pd.DataFrame.from_records(
                create_baseline_events(evs[evs.list>=1].to_records(), start_time=3000, end_time=30000)
            ).drop(columns='index')
        except IndexError:
            continue
        evs = evs[np.logical_or(evs.type=='REC_BASE', np.logical_and(evs.type=='REC_WORD', evs.intrusion==0))]
        rec_mask = (evs.type=='REC_WORD').values.astype(bool)
        if rec_mask.sum() == 0:
            cnt += 1
            continue
        pairs = reader.load('pairs')
        if not np.isin(hippo_pairs, pairs.label).all():
            # localization does not match
            continue
        good_pairs = pairs[np.isin(pairs.label, hippo_pairs)]
        # resample so that number of frequencies is the same
        eeg = reader.load_eeg(evs, scheme = good_pairs, rel_start=settings.rel_start, rel_stop=settings.rel_stop)
        eeg = eeg.resample(500.)
        eeg = eeg.to_ptsa()
        if not len(full_eeg):
            full_eeg.append(eeg)
        else:
            for j in range(len(full_eeg)):
                eeg, full_eeg[j] = trim_time_ts(eeg, full_eeg[j])
            full_eeg.append(eeg)
        events.append(evs)
    full_eeg = concat(full_eeg, 'event')
    events = pd.concat(events)
    events['item_name'] = events['item_name'].astype(str)
    events['recalled'] = events['type']=="REC_WORD"
    bt_filt = ButterworthFilter(freq_range=[58., 62.])
    full_eeg = full_eeg.filter_with(bt_filt)
    irasa = IRASA(full_eeg, settings.freqs, samplerate = full_eeg.samplerate.values, hset=settings.hset)
    p, plaw = irasa.plaw_fit()
    osc = irasa.mixed - irasa.fractal
    osc = SSL_transform(osc)
    mixed = np.log10(irasa.mixed)
    fractal = np.log10(irasa.fractal)
    # z score over events (axis=0)
    ts = TimeSeries(np.stack([osc, fractal, mixed]),
           dims=('type', 'event', 'channel', 'frequency'),
           coords = {'type':['oscillatory', 'fractal', 'mixed'],
                     'event':pd.MultiIndex.from_frame(events[["eegoffset","session", "item_name", "recalled"]]),
                     'frequency':settings.freqs,
                     'channel':full_eeg.channel.values, 'samplerate':full_eeg.samplerate.values}
          )
    ts = ts.groupby('session').reduce(zscore, dim='event', ddof=1)
    
    if cnt == len(data):
        raise Exception('Subject has no recalled events')
    if settings.save:
        ts.to_hdf(settings.data_dir + f"{sub}_bip_irasa.h5")
    return ts

def calc_avg_oscill_rec(sub, settings_path='settings_IRASA_retrieval_hmax_2.pkl'):
    import sys
    sys.path.append('/home1/jrudoler/src/')
    from retrieval_help import create_baseline_events, free_epochs
    settings = da.Settings.Load(filename=settings_path)
    subjects = pd.read_pickle(settings.subj_pairs_path)
    hippo_pairs = subjects[subjects.Subject==sub].hippo_pairs.iloc[0]
    hippo_pairs = np.ravel([i.split('-') for i in hippo_pairs])
    data = cml.get_data_index(kind = 'r1'); data = data[data['experiment'] =='FR1']; data = data[data['subject']==sub]
    full_eeg = []
    cnt = 0
    events = []
    for i, row in data.iterrows(): # iterate then average over sessions
        reader = cml.CMLReader(sub, 'FR1', session = row['session'], 
                               localization=row['localization'], montage=row['montage'])
        evs = reader.load('all_events')
        evs = evs[evs.eegoffset>0]
        # drop retrieval events where preceding word was < 1000 ms before
        rec_mask = (evs.type=='REC_WORD').values.astype(bool)
        evs = evs.drop(index=evs[rec_mask].index[(evs[rec_mask].mstime.diff()<1000)].values)
        # get matched baseline events to compare to recalls
        try:
            evs = pd.DataFrame.from_records(
                create_baseline_events(evs[evs.list>=1].to_records(), start_time=3000, end_time=30000)
            ).drop(columns='index')
        except IndexError:
            continue
        evs = evs[np.logical_or(evs.type=='REC_BASE', np.logical_and(evs.type=='REC_WORD', evs.intrusion==0))]
        rec_mask = (evs.type=='REC_WORD').values.astype(bool)
        if rec_mask.sum() == 0:
            cnt += 1
            continue
        pairs = reader.load('pairs')
        contacts = reader.load('contacts')
        if not np.isin(hippo_pairs, contacts.label).all():
            # localization does not match
            continue
        good_pairs = pairs[np.isin(pairs.label, hippo_pairs)]
        # resample so that number of frequencies is the same
        eeg = reader.load_eeg(evs, scheme = contacts, rel_start=settings.rel_start, rel_stop=settings.rel_stop)
        eeg = eeg.resample(500.)
        eeg = eeg.to_ptsa()
        # average reference
        good_contacts = np.isin(contacts.label, hippo_pairs)
        eeg = eeg[:, good_contacts, :] - eeg.mean('channel')
        if not len(full_eeg):
            full_eeg.append(eeg)
        else:
            for j in range(len(full_eeg)):
                eeg, full_eeg[j] = trim_time_ts(eeg, full_eeg[j])
            full_eeg.append(eeg)
        events.append(evs)
    full_eeg = concat(full_eeg, 'event')
    events = pd.concat(events)
    events['item_name'] = events['item_name'].astype(str)
    events['recalled'] = events['type']=="REC_WORD"
    bt_filt = ButterworthFilter(freq_range=[58., 62.])
    full_eeg = full_eeg.filter_with(bt_filt)
    irasa = IRASA(full_eeg, settings.freqs, samplerate = full_eeg.samplerate.values, hset=settings.hset)
    p, plaw = irasa.plaw_fit()
    osc = irasa.mixed - irasa.fractal
    osc = SSL_transform(osc)
    mixed = np.log10(irasa.mixed)
    fractal = np.log10(irasa.fractal)
    # z score over events (axis=0)
    ts = TimeSeries(np.stack([osc, fractal, mixed]),
           dims=('type', 'event', 'channel', 'frequency'),
           coords = {'type':['oscillatory', 'fractal', 'mixed'],
                     'event':pd.MultiIndex.from_frame(events[["eegoffset","session", "item_name", "recalled"]]),
                     'frequency':settings.freqs,
                     'channel':full_eeg.channel.values, 'samplerate':full_eeg.samplerate.values}
          )
    ts = ts.groupby('session').reduce(zscore, dim='event', ddof=1)
    
    if cnt == len(data):
        raise Exception('Subject has no recalled events')
    if settings.save:
        ts.to_hdf(settings.data_dir + f"{sub}_avg_irasa.h5")
    return ts

def calc_bip_morlet(sub, settings_path='settings_IRASA_hmax_2.pkl'):   
    settings = da.Settings.Load(filename=settings_path)
    subjects = pd.read_pickle(settings.subj_pairs_path)
    hippo_pairs = subjects[subjects.Subject==sub].hippo_pairs.iloc[0]
    data = cml.get_data_index(kind = 'r1'); data = data[data['experiment'] =='FR1']; data = data[data['subject']==sub]
    recalls = []
    full_eeg = []
    # iterate then average over sessions
    for i, row in data.reset_index().iterrows():
        reader = cml.CMLReader(sub, 'FR1', session = row['session'], 
                               localization=row['localization'], montage=row['montage'])
        events = reader.load('task_events')
        word_events= events[(events.type == 'WORD')&(events.eegoffset != -1)]
        word_events= word_events[['eegfile', 'eegoffset', 'experiment', 'intrusion', 'item_name',
                                  'item_num', 'list', 'montage', 'msoffset', 'mstime', 'recalled', 
                                  'rectime', 'serialpos', 'session', 'subject', 'type']]
        pairs = reader.load('pairs')
        if not np.isin(hippo_pairs, pairs.label).all():
            # localization does not match
            continue
        pairs = pairs[np.isin(pairs.label, hippo_pairs)]
        # resample so that number of frequencies is the same
        eeg = reader.load_eeg(word_events, scheme = pairs,
                              rel_start=settings.rel_start - 1000,
                              rel_stop=settings.rel_stop + 1000)
        eeg = eeg.to_ptsa()
        eeg = eeg.resampled(500.)
        if not len(full_eeg):
            full_eeg.append(eeg)
        else:
            for j in range(len(full_eeg)):
                eeg, full_eeg[j] = trim_time_ts(eeg, full_eeg[j])
            full_eeg.append(eeg)
        recalls.extend(word_events.recalled.values.astype(bool))
    if np.mean(recalls)<.08:
        # subject remembers less than one word per list
        raise Exception('Subject does not meet recall requirements')
    full_eeg = concat(full_eeg, 'event')
    recalls = full_eeg
    bt_filt = ButterworthFilter(freq_range=[58., 62.])
    mw_filt = MorletWaveletFilter(freqs=settings.freqs, width=4, output='power')
    powers = full_eeg.filter_with([bt_filt, mw_filt])
    powers = powers.remove_buffer(1.)
    # log transform
    powers = powers.reduce(np.log10)
    powers = powers.mean('time')
    powers = powers.transpose('event', 'channel', 'frequency')
    # z score over events (axis=0)
    powers = powers.groupby("session").reduce(zscore, ddof=1, dim='event')
    if settings.save:
        if "stim_params" in powers.indexes['event'].names:
            powers = powers.assign_coords({'event':powers.indexes['event'].droplevel('stim_params')})
        powers.to_hdf(settings.data_dir+'{}_bip_morlet.h5'.format(sub))
    return powers

## retrieval
def calc_bip_morlet_rec(sub, settings_path='settings_IRASA_retrieval_hmax_2.pkl'):
    import sys
    sys.path.append('/home1/jrudoler/src/')
    from retrieval_help import create_baseline_events, free_epochs
    settings = da.Settings.Load(filename=settings_path)
    subjects = pd.read_pickle(settings.subj_pairs_path)
    hippo_pairs = subjects[subjects.Subject==sub].hippo_pairs.iloc[0]
    data = cml.get_data_index(kind = 'r1'); data = data[data['experiment'] =='FR1']; data = data[data['subject']==sub]
    full_eeg = []
    cnt = 0
    events = []
    for i, row in data.iterrows(): # iterate then average over sessions
        reader = cml.CMLReader(sub, 'FR1', session = row['session'], 
                               localization=row['localization'], montage=row['montage'])
        evs = reader.load('all_events')
        evs = evs[evs.eegoffset>0]
        evs = evs[['eegfile', 'eegoffset', 'experiment', 'intrusion', 'item_name',
                                  'item_num', 'list', 'montage', 'msoffset', 'mstime', 'recalled', 
                                  'rectime', 'serialpos', 'session', 'subject', 'type']]
        # drop retrieval events where preceding word was < 1000 ms before
        rec_mask = (evs.type=='REC_WORD').values.astype(bool)
        evs = evs.drop(index=evs[rec_mask].index[(evs[rec_mask].mstime.diff()<1000)].values)
        # get matched baseline events to compare to recalls
        try:
            evs = pd.DataFrame.from_records(
                create_baseline_events(evs[evs.list>=1].to_records(), start_time=3000, end_time=30000)
            ).drop(columns='index')
        except IndexError:
            continue
        evs = evs[np.logical_or(evs.type=='REC_BASE', np.logical_and(evs.type=='REC_WORD', evs.intrusion==0))]
        rec_mask = (evs.type=='REC_WORD').values.astype(bool)
        if rec_mask.sum() == 0:
            print("no recalled words")
            cnt += 1
            continue
        pairs = reader.load('pairs')
        if not np.isin(hippo_pairs, pairs.label).all():
            continue
        pairs = pairs[np.isin(pairs.label, hippo_pairs)]
        # resample so that number of frequencies is the same
        eeg = reader.load_eeg(evs, scheme = pairs,
                              rel_start=settings.rel_start - 1000*settings.buffer_sec,# - 1000,
                              rel_stop=settings.rel_stop)#  + 1000)
        eeg = eeg.resample(500.)
        eeg = eeg.to_ptsa()
        # mirrored buffer for retrieval
        eeg = eeg.add_mirror_buffer(settings.buffer_sec, two_sided=False)
        if not len(full_eeg):
            full_eeg.append(eeg)
        else:
            for j in range(len(full_eeg)):
                eeg, full_eeg[j] = trim_time_ts(eeg, full_eeg[j])
            full_eeg.append(eeg)
        events.append(evs)
    full_eeg = concat(full_eeg, 'event')
    events = pd.concat(events)
    events['item_name'] = events['item_name'].astype(str)
    events['recalled'] = events['type']=="REC_WORD"
    bt_filt = ButterworthFilter(freq_range=[58., 62.])
    mw_filt = MorletWaveletFilter(freqs=settings.freqs, width=4, output='power')
    powers = full_eeg.filter_with([bt_filt, mw_filt])
    powers = powers.remove_buffer(settings.buffer_sec)
    # log transform
    powers = powers.reduce(np.log10)
    powers = powers.mean('time')
    powers = powers.transpose('event', 'channel', 'frequency')
    # z score over events (axis=0)
    powers = powers.reduce(zscore, ddof=1, dim='event')
    # make sure no funny business
    powers = powers.assign_coords({'event':pd.MultiIndex.from_frame(events)})
    if cnt == len(data):
        raise Exception('Subject has no recalled events')
    if settings.save:
        powers.to_hdf(settings.data_dir+'{}_bip_morlet.h5'.format(sub))
    return powers